import argparse
import functools
import textwrap
import gzip
import time
import os
import re


class Singleton(type):
    _instances = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super(Singleton, cls).__call__(*args, **kwargs)
        return cls._instances[cls]


def report_execution_time(step_name: str):
    def timer(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            start = time.time()
            return_val = func(*args, **kwargs)
            end = time.time()
            hrs, rem = divmod(end - start, 3600)
            min, sec = divmod(rem, 60)
            units = [(hrs, '%dh'), (min, '%dm'), (sec, '%.2fs')]
            print(f"{step_name} took {' '.join([u % q for q, u in units if q > 0])}")
            return return_val
        return wrapper
    return timer


class SmartFormatter(argparse.HelpFormatter):
    # Properly formats argparse helpstring fields that contain newlines
    def _split_lines(self, text, width):
        if text.startswith('R|'):
            lines = []
            for line in text[2:].splitlines():
                fill = textwrap.fill(line, width, subsequent_indent="  ")
                lines.extend(fill.splitlines())
            return lines
        return argparse.HelpFormatter._split_lines(self, text, width)


def from_here(config_file, input_file):
    """Calculates paths relative to the config file which contains them"""
    config_file, input_file = (os.path.expanduser(p) for p in [config_file, input_file])

    if not os.path.isabs(input_file):
        from_here = os.path.dirname(config_file)
        input_file = os.path.normpath(os.path.join(from_here, input_file))

    return input_file


def make_filename(args, ext='.csv'):
    return '_'.join([str(chnk) for chnk in args if chnk is not None]) + ext


def get_r_safename(name: str) -> str:
    """Converts a string to a syntactically valid R name

    This can be used to match names along axes of DataFrames produced by R,
    assuming that the R script takes no measures to preserve names itself.
    https://stat.ethz.ch/R-manual/R-devel/library/base/html/make.names.html
    """

    leading_char = lambda x: re.sub(r"^(?=[^a-zA-Z.]+|\.\d)", "X", x)
    special_char = lambda x: re.sub(r"[^a-zA-Z0-9_.]", ".", x)
    return special_char(leading_char(name))


class ReadOnlyDict(dict):
    """A very simple "read-only" wrapper for dictionaries. This will primarily be used
    for passing around argparse's command line args as a dictionary, but ensuring that
    preferences cannot be accidentally modified as they are passed around to a menagerie
    of class constructors. All methods except __setitem__() are deferred to the base class."""

    def __init__(self, rw_dict):
        super().__init__(rw_dict)

    def __setitem__(self, *_):
        raise RuntimeError("Attempted to modify read-only dictionary after construction.")

def sorted_natural(lines, reverse=False):
    """Sorts alphanumeric strings with entire numbers considered in the sorting order,
    rather than the default behavior which is to sort by the individual ASCII values
    of the given number. Returns a sorted copy of the list, just like sorted().

    Not sure who to credit... it seems this snippet has been floating around for quite
    some time. Strange that there isn't something in the standard library for this."""

    convert = lambda text: int(text) if text.isdigit() else text.lower()
    alphanum_key = lambda key: [convert(c) for c in re.split(r'(\d+)', key)]
    return sorted(lines, key=alphanum_key, reverse=reverse)


# File IO interface for reading and writing Gzip files
gzip_open = functools.partial(gzip.GzipFile, compresslevel=6, fileobj=None, mtime=0)